# -*- coding: utf-8 -*-

# Copyright 2008 Jaap Karssenberg <pardus@cpan.org>

'''This package contains third party libraries to be packaged
with zim. See the respective files for the copyright and license
notices.
'''
